/*
 * Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
 *
 * This file is part of NRAO Workspaces.
 *
 * Workspaces is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Workspaces is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
 */
import { Component, OnInit } from "@angular/core";
import { Capability } from "../../model/capability";
import { CapabilityRequest } from "../../model/capability-request";
import { CapabilityRequestService } from "../../services/capability-request.service";
import { CapabilityLauncherService } from "../../services/capability-launcher.service";
import { DataRetrieverService } from "../../services/data-retriever.service";
import { PollingDataUpdaterService } from "../../services/polling-data-updater.service";
import { Subject } from "rxjs";
import { takeUntil } from "rxjs/operators";

@Component({
  selector: "app-list-of-calibratables",
  templateUrl: "./list-of-calibratables.component.html",
  styleUrls: ["./list-of-calibratables.component.scss"],
})
export class ListOfCalibratablesComponent implements OnInit {
  public calibratableRequests: Array<CapabilityRequest>;
  public capability: Capability;
  public isPaused: boolean;
  private stdCalibrationCapability = "std_calibration";

  private ngUnsubscribe = new Subject(); // to signal when all subscribers should unsubscribe (triggered in tear down)

  constructor(
    private capabilityRequestService: CapabilityRequestService,
    private capabilityLauncherService: CapabilityLauncherService,
    private pollingDataUpdaterService: PollingDataUpdaterService,
    private dataRetriever: DataRetrieverService,
  ) {
    this.pollingDataUpdaterService
      .getDataPoller$(this.dataRetriever.getCapabilityUrl(this.stdCalibrationCapability))
      .pipe(takeUntil(this.ngUnsubscribe))
      .subscribe(this.capabilityObserver);
    this.pollingDataUpdaterService
      .getDataPoller$(
        this.capabilityRequestService.getActiveCapabilityRequestsUrl(this.stdCalibrationCapability),
      )
      .pipe(takeUntil(this.ngUnsubscribe))
      .subscribe(this.createdRequestsObserver);
  }

  ngOnInit(): void {}

  private createdRequestsObserver = {
    next: (request) => {
      this.calibratableRequests = request.active_requests;
    },
    error: (error) =>
      console.error("Error when retrieving list of created capability requests:" + error),
  };

  // Observer for capability objects
  public capabilityObserver = {
    next: (capability) => {
      this.capability = capability;
      this.isPaused = this.capability.paused;
    },
    error: (error) => console.error("Error when retrieving capability:", error),
  };

  /**
   * Get metadata from capability request parameters if it exists
   * @param request Capability request
   * @return metadata Metadata section of parameters
   */
  public getMetadata(request: CapabilityRequest): any {
    if (request.parameters.hasOwnProperty("metadata")) {
      return request.parameters.metadata.valueOf();
    } else {
      return {};
    }
  }

  public togglePauseCapability(): void {
    const togglePauseCapabilityObservable = {
      next: (request) => {},
      error: (error) => {
        console.error("Error when pausing capability:" + error);
      },
    };

    let pauseAction = this.isPaused ? "unpause" : "pause";
    this.dataRetriever
      .togglePauseCapability(this.stdCalibrationCapability, pauseAction)
      .subscribe(togglePauseCapabilityObservable);
  }
}
